-- Provides extended table functionality
Tables = {}

-- Counts the number of elements in a table
Tables.CountTable = function (t)
    if not t then return 0 end
    if (type(t) ~= "table") then return 0 end
    local count = 0
    for key, value in pairs(t) do
        count = count + 1
    end
    return count
end
 
-- Returns a bool indicating if the input table contains the
-- input element. Returns false if either input is invalid.
-- Optional 'compare' function allows for custom match logic.
Tables.Contains = function (t, element, CompareFunction)
	if not t or not element then return false end
	
	if not CompareFunction then
		CompareFunction = Tables.DefaultCompareFunction
	end
	
	for _, value in ipairs(t) do
		if CompareFunction(element, value) then return true end
	end
	return false
end

-- The default method used to compare two elements of a table
Tables.DefaultCompareFunction = function (element1, element2)
	return element1 == element2
end
 
-- Return all elements from the 't' table excluding any elements that are also
-- in the 'exclude' table. Allow an optional custom comparison function as an input.
Tables.Exclude = function (t, exclude, CompareFunction)
    local output = {}
	if not CompareFunction then
		CompareFunction = Tables.DefaultCompareFunction
	end

    -- Iterate over the first table and add elements not in the second table to the result
    for key, element in pairs(t) do
		if key and element then
			if not Tables:Contains(exclude, element, CompareFunction) then
				output[key] = element
			end
		end
    end
    return output
end

-- Accepts a single table as inputs; returns a single table with
-- every unique value in the input table appearing a single time
-- Optional comparison function input allows custom logic.
Tables.Unique = function (t, CompareFunction)
    local output = {}
	if not CompareFunction then
		CompareFunction = Tables.DefaultCompareFunction
	end

	for key, element in pairs(t) do
		if key and element then
			if not Tables:Contains(output, element, CompareFunction) then
				output[key] = element
			end
		end
	end
    return output
end

-- Take two tables as inputs, return the intersection
Tables.Intersection = function (t1, t2)
    local intersection = {}
    local set = {}

    -- Add elements from the first table to the set
    for key, value in pairs(t1) do
        set[value] = true
    end

    -- Check elements in the second table
    for key, value in pairs(t2) do
        if set[value] then
            table.insert(intersection, value)
            set[value] = nil -- To avoid duplicate entries in the intersection
        end
    end

    return intersection
end

-- Returns a random value from the input table
-- Nil if any error
Tables.GetRandomValue = function (t)
	if not t then return nil end
	if #t == 0 then return nil end

	-- Collect valid abilities into a list
	local keys = {}
	for key in pairs(t) do
		table.insert(keys, key)
	end

	-- Get a random value from the list
	local randomkey = keys[math.random(#keys)]
	return t[randomkey]
end

-- Returns the Nth value of the input table
-- Returns nil if anything goes wrong 
Tables.GetValue = function (t, n)
    if not t or type(t) ~= "table" then return nil end
    local count = 0
    for _, value in pairs(t) do
        count = count + 1
        if count == n then
            return value
        end
    end
    return nil
end

-- Returns a copy of the input table skipping the first element
-- Agnostic to key type and order
Tables.ExcludeFirst = function(t)
    if not t then return {} end
    local output = {}
    local index = 1

    for k, v in pairs(t) do
        if type(k) == "number" and k > 1 then
            output[index] = v
            index = index + 1
        elseif type(k) ~= "number" then
            output[k] = v
        end
    end

    return output
end

-- Copy a table and return the new object
Tables.CopyTable = function(t)
    local output = {}
    for key, value in pairs(t) do
        if type(value) == "table" then
            output[key] = Tables.CopyTable(value)
        else
            output[key] = value
        end
    end
    return output
end


-- Sort the table t by key using an input comparator function
Tables.SortByKeys = function (t, comparator)
    local keys = {}
    for key in pairs(t) do table.insert(keys, key) end
    table.sort(keys, function(a, b) return comparator(a) < comparator(b) end)

    local output = {}
    for _, key in ipairs(keys) do output[key] = t[key] end
    return output
end

-- Sort the table t by value using an input comparator function
Tables.SortByValues = function(t, comparator)
    local values = {}
    for key, value in pairs(t) do 
        table.insert(values, {key = key, value = value}) 
    end
    
    table.sort(values, function(a, b) 
        return comparator(a.value, b.value) < 0 
    end)

    local output = {}
    for _, entry in ipairs(values) do 
        output[entry.key] = entry.value 
    end
    return output
end